/*
 * File descriptor interface.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>

#include <sys/types.h>

EXTERN_C_BEGIN

/* Simple shared buffer state. */

typedef struct
{
  /* Shared buffer communication. */

  char *addr;           /* shared buffer address */
  size_t size;          /* size of the buffer */

} fs_buffer_t;

/* Common filesystem object state. */

typedef struct
{
  /* Shared between client and server by value exchange. */

  size_t start_pos;     /* start position of the mapped dataspace in the file */
  size_t end_pos;       /* end position of the mapped dataspace in the file */
  size_t size;          /* file size (used by the client for seek) or
                           region size (for pipes, identical to buffer.size) */

  /* Shared buffer communication. */

  fs_buffer_t buffer;

} fs_object_t;

/* Descriptor for filesystem objects. */

typedef struct
{
  l4_cap_idx_t server;  /* filesystem server */

  l4_cap_idx_t context; /* opener context reference
                           (also acting as dataspace) */
  l4_cap_idx_t ref;     /* filesystem object reference
                           (also acting as dataspace) */

  fs_buffer_t cobj;     /* context object state */
  fs_object_t obj;      /* common filesystem object state */

  /* Configuration. */

  int flags;            /* opening flags */

  /* Client-side operations. */

  int written;          /* indicates whether data has been written */

  /* Client-side dataspace offsets. */

  size_t data_start;    /* offset of available data */
  size_t data_end;      /* offset of the end of available data */

  /* File/pipe indicator. */

  int can_mmap;         /* indicates whether mmap can be used instead of region
                           navigation */

} file_descriptor_t;

/* Allocation operations. */

file_descriptor_t *fsdesc_init_cap(file_descriptor_t *desc, const char *cap);

/* Deallocation operations. */

void fsdesc_deallocate(file_descriptor_t *desc);
void fsdesc_deallocate_buffer(fs_buffer_t *buffer);
void fsdesc_deallocate_caps(file_descriptor_t *desc);
void fsdesc_deallocate_context(file_descriptor_t *desc);

/* IPC-related initialisation. */

char *fsdesc_get_name(fs_buffer_t *buffer);
void fsdesc_set_name(fs_buffer_t *buffer, const char *name);
long fsdesc_init_buffer_from_dataspace(fs_buffer_t *buffer, l4_cap_idx_t ds);
void fsdesc_init_buffer(fs_buffer_t *buffer, char *addr, unsigned long size);

/* State operations. */

void fsdesc_reset_buffer(file_descriptor_t *desc, off_t position);

EXTERN_C_END
