/*
 * File descriptor allocation.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <string.h>
#include <sys/types.h>

#include <l4/re/env.h>
#include <l4/sys/err.h>
#include <l4/sys/types.h>

#include "fsdesc.h"
#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>



/* Client operations. */

/* Initialise an object for interaction with the filesystem. */

file_descriptor_t *fsdesc_init_cap(file_descriptor_t *desc, const char *cap)
{
  /* Set referenced capabilities to invalid, initially. */

  desc->ref = L4_INVALID_CAP;
  desc->context = L4_INVALID_CAP;

  /* Allocate a capability for the server. */

  desc->server = l4re_env_get_cap(cap);

  if (l4_is_invalid_cap(desc->server))
    return NULL;

  /* Indicate an absent memory region. */

  desc->obj.buffer.addr = 0;
  desc->obj.buffer.size = 0;

  /* Initialise state independently of the server. */

  fsdesc_reset_buffer(desc, 0);
  return desc;
}



/* Buffer access functions. */

char *fsdesc_get_name(fs_buffer_t *buffer)
{
  /* Reject unterminated strings. */

  if (strnlen(buffer->addr, buffer->size) < buffer->size)
    return buffer->addr;
  else
    return NULL;
}

void fsdesc_set_name(fs_buffer_t *buffer, const char *name)
{
  /* Truncate the string to the buffer size (minus terminator). */
  /* NOTE: Would use strncpy here. */

  unsigned long i;

  for (i = 0; i < buffer->size - 1; i++)
  {
    buffer->addr[i] = name[i];
    if (!name[i])
      break;
  }

  buffer->addr[buffer->size - 1] = '\0';
}



/* Reset the buffer to an empty state. */

void fsdesc_reset_buffer(file_descriptor_t *desc, off_t position)
{
  desc->data_start = 0;
  desc->data_end = 0;
  desc->written = 0;
  desc->can_mmap = 0;
  desc->obj.size = 0;
  desc->obj.start_pos = position;
}



/* Deallocate a filesystem object and associated resources. */

void fsdesc_deallocate(file_descriptor_t *desc)
{
  if (desc == NULL)
    return;

  fsdesc_deallocate_caps(desc);
  fsdesc_deallocate_context(desc);
}

/* Deallocate server and filesystem object references. */

void fsdesc_deallocate_caps(file_descriptor_t *desc)
{
  if (l4_is_valid_cap(desc->server))
  {
    ipc_cap_free_um(desc->server);
    desc->server = L4_INVALID_CAP;
  }

  if (l4_is_valid_cap(desc->ref))
  {
    fsdesc_deallocate_buffer(&desc->obj.buffer);
    ipc_cap_free_um(desc->ref);
    desc->ref = L4_INVALID_CAP;
  }
}

/* Deallocate context reference. */

void fsdesc_deallocate_context(file_descriptor_t *desc)
{
  if (l4_is_valid_cap(desc->context))
  {
    fsdesc_deallocate_buffer(&desc->cobj);
    ipc_cap_free_um(desc->context);
    desc->context = L4_INVALID_CAP;
  }
}

/* Deallocate dataspace-related buffers. */

void fsdesc_deallocate_buffer(fs_buffer_t *buffer)
{
  if (buffer->addr)
  {
    ipc_detach_dataspace(buffer->addr);
    buffer->addr = 0;
  }
}



/* Initialise a filesystem object structure with the given dataspace. */

long fsdesc_init_buffer_from_dataspace(fs_buffer_t *buffer, l4_cap_idx_t ds)
{
  unsigned long size;

  /* NOTE: Truncation of dataspace size to word-sized member. */

  long err = ipc_dataspace_size(ds, &size);
  if (err)
    return err;

  buffer->size = size;

  /* Attach the dataspace. */

  return ipc_attach_dataspace(ds, buffer->size, (void **) &buffer->addr);
}

/* Initialise a filesystem object structure with buffer details. */

void fsdesc_init_buffer(fs_buffer_t *buffer, char *addr, unsigned long size)
{
  buffer->addr = addr;
  buffer->size = size;
}
